const rp = require('request-promise-native');
const crypto = require('crypto');

function wxSDK(appID, secret, token, AppKey, db) {
    if (typeof appID == 'object') {
        if (!appID.appID) throw 'wxSDK初始化错误,appID不能为空!';
        if (!appID.secret) throw 'wxSDK初始化错误,secret不能为空!';
        if (!appID.token) throw 'wxSDK初始化错误,token不能为空!';
        if (!appID.AppKey) throw 'wxSDK初始化错误,AppKey不能为空!';

        this.appID = appID.appID;
        this.secret = appID.secret;
        this.token = appID.token;
        this.aesKey = new Buffer(appID.AppKey + '=', 'base64');
        this.iv = this.aesKey.slice(0, 16);
        this.db = secret;
    } else {
        if (!appID) throw 'wxSDK初始化错误,appID不能为空!';
        if (!secret) throw 'wxSDK初始化错误,secret不能为空!';
        if (!token) throw 'wxSDK初始化错误,token不能为空!';
        if (!AppKey) throw 'wxSDK初始化错误,AppKey不能为空!';
        this.appID = appID;
        this.secret = secret;
        this.token = token;
        this.aesKey = new Buffer(AppKey + '=', 'base64');
        this.iv = this.aesKey.slice(0, 16);
        this.db = db;
    }
}
async function post(url, body, name) {
    let resultjson = await rp.post({
        url: url,
        method: 'POST',
        rejectUnauthorized: false,  //启用拒绝未授权证书
        body: body,
        json: true
    });
    global.app.wxlogger.info('========post', name, JSON.stringify(body));
    global.app.wxlogger.info('========post', name, resultjson);
    if (resultjson.errcode) throw new Error(name + '时发生错误(' + resultjson.errcode + '):' + resultjson.errmsg || JSON.stringify(resultjson));

    return resultjson;
}
async function get(url, name) {
    let resultjson = await rp.get({
        url: url,
        method: 'GET',
        json: true
    });
    global.app.wxlogger.info('========get', name, resultjson);
    if (resultjson.errcode) throw new Error(name + '时发生错误(' + resultjson.errcode + '):' + resultjson.errmsg || JSON.stringify(resultjson));

    return resultjson;
}
wxSDK.prototype.获取第三方平台预授权 = async function () {
    const url = 'https://api.weixin.qq.com/cgi-bin/component/api_create_preauthcode?component_access_token=' + await this.获取第三方平台Token();

    return await post(url, {
        component_appid: this.appID,
    }, '获取第三方平台预授权');
};
wxSDK.prototype.获取公众号信息 = async function (authorizer_appid) {
    const url = 'https://api.weixin.qq.com/cgi-bin/component/api_get_authorizer_info?component_access_token=' + await this.获取第三方平台Token();

    return await post(url, {
        component_appid: this.appID,
        authorizer_appid: authorizer_appid
    }, '获取公众号信息');
};
wxSDK.prototype.公众号_菜单_查询 = async function (authorizer_appid, authorizer_refresh_token) {
    const url = 'https://api.weixin.qq.com/cgi-bin/menu/get?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return get(url, '公众号_菜单_查询');
};
wxSDK.prototype.公众号_菜单_创建 = async function (authorizer_appid, authorizer_refresh_token, jo_menu) {
    const url = 'https://api.weixin.qq.com/cgi-bin/menu/create?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, jo_menu, '公众号_菜单_创建');
};
wxSDK.prototype.公众号_发送消息 = async function (authorizer_appid, authorizer_refresh_token, postjson) {
    const url = 'https://api.weixin.qq.com/cgi-bin/message/custom/send?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, postjson, '公众号_发送消息');
};
//----------------------------
wxSDK.prototype.会员卡_创建类型 = async function (authorizer_appid, authorizer_refresh_token, postjson) {
    const url = 'https://api.weixin.qq.com/card/create?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, postjson, '会员卡_创建类型');
};
wxSDK.prototype.会员卡_修改卡类型 = async function (authorizer_appid, authorizer_refresh_token, postjson) {
    const url = 'https://api.weixin.qq.com/card/update?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, postjson, '会员卡_修改卡类型');
};
wxSDK.prototype.会员卡_设置开卡字段 = async function (authorizer_appid, authorizer_refresh_token, postjson) {
    const url = 'https://api.weixin.qq.com/card/membercard/activateuserform/set?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, postjson, '会员卡_设置开卡字段');
};
wxSDK.prototype.会员卡_删除卡类型 = async function (authorizer_appid, authorizer_refresh_token, card_id) {
    const url = 'https://api.weixin.qq.com/card/delete?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        card_id: card_id,
    }, '会员卡_删除卡类型');
};
wxSDK.prototype.会员卡_设置支付后投放 = async function (authorizer_appid, authorizer_refresh_token, postjson) {
    const url = 'https://api.weixin.qq.com/card/paygiftcard/add?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, postjson, '会员卡_设置支付后投放');
};
wxSDK.prototype.会员卡_删除支付后投放 = async function (authorizer_appid, authorizer_refresh_token, rule_id) {
    const url = 'https://api.weixin.qq.com/card/paygiftcard/delete?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        rule_id: rule_id,
    }, '会员卡_删除支付后投放');
};
wxSDK.prototype.会员卡_更新会员信息 = async function (authorizer_appid, authorizer_refresh_token, code, card_id, 卡面图片, 积分, 自定义1, 自定义2, 自定义3) {
    const url = 'https://api.weixin.qq.com/card/membercard/updateuser?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    let postjson = {
        code: code,
        card_id: card_id,
        notify_optional: {
            is_notify_custom_field1: true,
            is_notify_custom_field2: true,
            is_notify_custom_field3: true,
        },
    };
    if (卡面图片) postjson['background_pic_url'] = 卡面图片;
    if (积分) postjson['bonus'] = 积分;
    if (自定义1) postjson['custom_field_value1'] = 自定义1;
    if (自定义2) postjson['custom_field_value2'] = 自定义2;
    if (自定义3) postjson['custom_field_value3'] = 自定义3;

    return await post(url, postjson, '会员卡_更新会员信息');
};
wxSDK.prototype.会员卡_激活会员卡 = async function (authorizer_appid, authorizer_refresh_token, code, membership_number) {
    const url = 'https://api.weixin.qq.com/card/membercard/activate?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        code: code,
        membership_number: membership_number,
    }, '会员卡_激活会员卡');
};
wxSDK.prototype.会员卡_获取用户已领取卡 = async function (authorizer_appid, authorizer_refresh_token, openid, card_id) {
    const url = 'https://api.weixin.qq.com/card/user/getcardlist?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        openid: openid,
        card_id: card_id,
    }, '会员卡_获取用户已领取卡');
};
wxSDK.prototype.会员卡_获取用户提交资料 = async function (authorizer_appid, authorizer_refresh_token, activate_ticket) {
    const url = 'https://api.weixin.qq.com/card/membercard/activatetempinfo/get?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        activate_ticket: activate_ticket,
    }, '会员卡_获取用户提交资料');
};
wxSDK.prototype.encryptcode解码 = async function (authorizer_appid, authorizer_refresh_token, encrypt_code) {
    const url = 'https://api.weixin.qq.com/card/code/decrypt?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token);
    return await post(url, {
        encrypt_code: encrypt_code,
    }, 'encryptcode解码');
};

//-----------------------------
wxSDK.prototype.oauth2 = async function (authorizer_appid, code, component_appid) {
    const url = 'https://api.weixin.qq.com/sns/oauth2/component/access_token?appid=' + authorizer_appid + '&code=' + code + '&grant_type=authorization_code&component_appid=' + component_appid + '&component_access_token=' + await this.获取第三方平台Token();
    return await get(url, 'oauth2授权');
};
wxSDK.prototype.公众号_jssdk_ticket = async function (authorizer_appid, authorizer_refresh_token) {
    const url = 'https://api.weixin.qq.com/cgi-bin/ticket/getticket?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token) + '&type=jsapi';

    let WXtoken = this.db.getRepository('g_微信token').createQueryBuilder();
    let tokenrow = await WXtoken.where('appid = :appid and type = :type', { appid: authorizer_appid, type: '公众号jsticket' }).getOne();
    if (tokenrow && (((new Date() - tokenrow.gettime) / 1000) < (tokenrow.expires - 60))) return tokenrow.token;

    let { ticket, expires_in } = await get(url, '公众号_jssdk_ticket');

    if (tokenrow) {
        await WXtoken.update().set({
            'type': '公众号jsticket',
            appid: authorizer_appid,
            token: ticket,
            gettime: new Date(),
            expires: expires_in,
            refresh_token: '',
        }).execute();
    } else {
        await WXtoken.insert().values({
            'type': '公众号jsticket',
            appid: authorizer_appid,
            token: ticket,
            gettime: new Date(),
            expires: expires_in,
            refresh_token: '',
        }).execute();
    }
    return ticket;
};
wxSDK.prototype.公众号_jssdk_Sign = async function (jsapiTicket, url) {
    let nonceStr = parseInt((Math.random() * 100000000000), 10).toString();
    let timestamp = Date.now().toString();
    let str = 'jsapi_ticket=' + jsapiTicket + '&noncestr=' + nonceStr + '&timestamp=' + timestamp + '&url=' + url;

    const sha1 = crypto.createHash('sha1');
    sha1.update(str);
    let signature = sha1.digest('hex');

    return {
        url: url,
        jsapi_ticket: jsapiTicket,
        noncestr: nonceStr,
        timestamp: timestamp,
        signature: signature,
    };
};
wxSDK.prototype.公众号_卡券_ticket = async function (authorizer_appid, authorizer_refresh_token) {
    const url = 'https://api.weixin.qq.com/cgi-bin/ticket/getticket?access_token=' + await this.刷新公众号token(authorizer_appid, authorizer_refresh_token) + '&type=wx_card';

    let WXtoken = this.db.getRepository('g_微信token').createQueryBuilder();
    let tokenrow = await WXtoken.where('appid = :appid and type = :type', { appid: this.appID, type: '卡券apiticket' }).getOne();
    if (tokenrow && (((new Date() - tokenrow.gettime) / 1000) < (tokenrow.expires - 60))) return tokenrow.token;

    let { ticket, expires_in } = await get(url, '公众号_卡券_ticket');

    if (tokenrow) {
        await WXtoken.update().set({
            'type': '卡券apiticket',
            appid: this.appID,
            token: ticket,
            gettime: new Date(),
            expires: expires_in,
            refresh_token: '',
        }).execute();
    } else {
        await WXtoken.insert().values({
            'type': '卡券apiticket',
            appid: this.appID,
            token: ticket,
            gettime: new Date(),
            expires: expires_in,
            refresh_token: '',
        }).execute();
    }
    return ticket;
};
wxSDK.prototype.公众号_卡券_Sign = async function (api_ticket, card_id, code, openid) {
    let nonceStr = parseInt((Math.random() * 100000000000), 10).toString();
    let timestamp = parseInt(Date.now()/1000).toString();  //注意timestamp为秒
    let 数组 = [api_ticket, timestamp, card_id, openid, nonceStr];
    if (code) 数组.push(code);
    const str = 数组.sort().join('');

    const sha1 = crypto.createHash('sha1');
    sha1.update(str);
    let signature = sha1.digest('hex');

    return {
        openid: openid,
        timestamp: timestamp,
        nonce_str: nonceStr,
        signature: signature,
        // card_id: card_id,
        // code: code,
        // api_ticket: api_ticket,
    };
};

wxSDK.prototype.刷新公众号token = async function (authorizer_appid, authorizer_refresh_token) {
    const url = 'https://api.weixin.qq.com/cgi-bin/component/api_authorizer_token?component_access_token=' + await this.获取第三方平台Token();

    let WXtoken = this.db.getRepository('g_微信token').createQueryBuilder();
    let tokenrow = await WXtoken.where('appid = :appid and type = :type', { appid: authorizer_appid, type: '公众号Token' }).getOne();
    if (tokenrow && (((new Date() - tokenrow.gettime) / 1000) < (tokenrow.expires - 600))) return tokenrow.token;

    let resultjson = await post(url, {
        component_appid: this.appID,
        authorizer_appid: authorizer_appid,
        authorizer_refresh_token: authorizer_refresh_token
    }, '刷新公众号token');

    if (tokenrow) {
        await WXtoken.update().set({
            'type': '公众号Token',
            appid: authorizer_appid,
            token: resultjson.authorizer_access_token,
            gettime: new Date(),
            expires: resultjson.expires_in,
            refresh_token: resultjson.authorizer_refresh_token
        }).execute();
    } else {
        await WXtoken.insert().values({
            'type': '公众号Token',
            appid: authorizer_appid,
            token: resultjson.authorizer_access_token,
            gettime: new Date(),
            expires: resultjson.expires_in,
            refresh_token: resultjson.authorizer_refresh_token
        }).execute();
    }

    return resultjson.authorizer_access_token;
};
wxSDK.prototype.获取公众号token = async function (auth_code) {
    const url = 'https://api.weixin.qq.com/cgi-bin/component/api_query_auth?component_access_token=' + await this.获取第三方平台Token();

    let resultjson = await post(url, {
        component_appid: this.appID,
        authorization_code: auth_code
    }, '获取公众号token');

    //保存到数据库
    let WXtoken = this.db.getRepository('g_微信token').createQueryBuilder();
    let tokenrow = await WXtoken.where('appid = :appid and type = :type', { appid: resultjson.authorization_info.authorizer_appid, type: '公众号Token' }).getOne();

    if (tokenrow) {
        await WXtoken.update().set({
            'type': '公众号Token',
            appid: resultjson.authorization_info.authorizer_appid,
            token: resultjson.authorization_info.authorizer_access_token,
            gettime: new Date(),
            expires: resultjson.authorization_info.expires_in,
            refresh_token: resultjson.authorization_info.authorizer_refresh_token
        }).execute();
    } else {
        await WXtoken.insert().values({
            'type': '公众号Token',
            appid: resultjson.authorization_info.authorizer_appid,
            token: resultjson.authorization_info.authorizer_access_token,
            gettime: new Date(),
            expires: resultjson.authorization_info.expires_in,
            refresh_token: resultjson.authorization_info.authorizer_refresh_token
        }).execute();
    }

    return resultjson;
};
wxSDK.prototype.获取第三方平台Token = async function () {
    const url = 'https://api.weixin.qq.com/cgi-bin/component/api_component_token';

    let WXtoken = this.db.getRepository('g_微信token').createQueryBuilder();
    let tokenrow = await WXtoken.where('appid = :appid and type = :type', { appid: this.appID, type: '第三方平台Token' }).getOne();
    if (tokenrow && (((new Date() - tokenrow.gettime) / 1000) < (tokenrow.expires - 600))) return tokenrow.token;

    let ticketrow = await this.db.getRepository('g_全局变量').createQueryBuilder().getOne();
    if (!ticketrow || !ticketrow.ticket) throw new Error('获取第三方平台token时发生错误,ticket为空');

    let resultjson = await post(url, {
        component_appid: this.appID,
        component_appsecret: this.secret,
        component_verify_ticket: ticketrow.ticket//第三方ticket()
    }, '获取第三方平台Token');

    if (tokenrow) {
        await WXtoken.update().set({
            'type': '第三方平台Token',
            appid: this.appID,
            token: resultjson.component_access_token,
            gettime: new Date(),
            expires: resultjson.expires_in,
            refresh_token: '',
        }).execute();
    } else {
        await WXtoken.insert().values({
            'type': '第三方平台Token',
            appid: this.appID,
            token: resultjson.component_access_token,
            gettime: new Date(),
            expires: resultjson.expires_in,
            refresh_token: '',
        }).execute();
    }
    return resultjson.component_access_token;
};

module.exports = wxSDK;